import logging
from typing import Any

import httpx

from fitrequest.auth import Auth
from fitrequest.method_config import MethodConfig
from fitrequest.response import Response
from fitrequest.utils import extract_method_params, extract_url_params

logger = logging.getLogger(__name__)


class Session:
    synchronous: httpx.Client
    asynchronous: httpx.AsyncClient

    def __init__(
        self,
        client_name: str,
        version: str,
        auth: Auth | dict | None = None,
        base_url: str | None = None,
        headers: dict | None = None,
        **kwargs,
    ) -> None:
        self.base_url = base_url
        self.raw_auth = auth
        self.default_headers = {'User-Agent': f'fitrequest.{client_name}.{version}'} | (headers or {})
        self.update(headers=self.default_headers, **kwargs)

    def authenticate(self) -> None:
        """
        Retrieve the necessary data for authentication
        and configure the auth attribute for both synchronous and asynchronous clients.
        """
        if self.raw_auth is None:
            return
        auth = Auth(**self.raw_auth) if isinstance(self.raw_auth, dict) else self.raw_auth

        # auth.authentication property retreives necessary data for authentication
        self.synchronous.auth = auth.authentication
        self.asynchronous.auth = auth.authentication

    def update(
        self,
        auth: Auth | dict | None = None,
        base_url: str | None = None,
        headers: dict | None = None,
        **kwargs,
    ) -> None:
        """
        Update the session used by generated methods.

        It effectively drops the existing synchronous and asynchronous sessions
        and creates a new ones using the specified arguments.

        - If ``base_url`` is not None, it takes precedence over the configured base_url in the `MethodConfig` class.
        - The ``headers`` argument is merged with the default headers provided by **fitrequest**.
        - Additional keyword arguments are passed directly to the ``httpx`` clients during initialization.

        This method only updates the ``raw_auth`` attributes and does not trigger a new authentication process,
        including retrieving secret data or setting the httpx client's auth attribute.
        If you need to perform a new authentication, use the ``authenticate`` method.
        """
        self.base_url = base_url
        self.raw_auth = auth or self.raw_auth

        self.config = {
            'headers': self.default_headers | (headers or {}),
            'follow_redirects': True,
            'timeout': 60,
        } | kwargs

        # Close previous sessions
        if hasattr(self, 'synchronous'):
            self.synchronous.close()

        if hasattr(self, 'asynchronous'):
            self.synchronous.close()

        # Create new session with updated parameters
        self.asynchronous = httpx.AsyncClient(transport=httpx.AsyncHTTPTransport(), **self.config)
        self.synchronous = httpx.Client(transport=httpx.HTTPTransport(), **self.config)

    def request(
        self,
        method_config: MethodConfig,
        raise_for_status: bool | None = None,
        filepath: str | None = None,
        **kwargs,
    ) -> Any:
        """
        Sends a request to a URL created based on the provided MethodConfig.

        Any keyword arguments (kwargs) you provide will be used as named variables to build the final URL,
        and some of these kwargs are also passed directly to the httpx.request method.

        You can optionally supply a custom URL, which will take priority over the generated one.
        If a custom URL with hardcoded parameters is provided along with the argument "params" (in kwargs),
        both sets of parameters are combined, giving priority to those in the custom URL over the ones in "params."
        """

        if raise_for_status is None:
            raise_for_status = method_config.raise_for_status

        logger.debug(
            'Request url',
            extra={
                'method_config': method_config.model_dump(exclude_none=True),
                'raise_for_status': raise_for_status,
                'kwargs': kwargs,
            },
        )

        custom_params = {key: val for key, val in kwargs.pop('params', {}).items() if val is not None}
        custom_url, custom_url_params = extract_url_params(kwargs.pop('url', None))
        custom_method = kwargs.pop('method', None)

        if self.base_url:
            method_config.base_url = self.base_url

        httpx_response = self.synchronous.request(
            method=custom_method or method_config.request_verb.value,
            url=custom_url or method_config.url(**kwargs),
            params=custom_params | custom_url_params or None,
            **extract_method_params(httpx.request, kwargs),
        )
        response = Response(
            httpx_response=httpx_response,
            json_path=method_config.json_path,
            raise_for_status=raise_for_status,
            response_model=method_config.response_model,
        )

        if filepath and method_config.save_method:
            response.save_data(filepath)
            return None
        return response.data

    async def async_request(
        self,
        method_config: MethodConfig,
        raise_for_status: bool | None = None,
        filepath: str | None = None,
        **kwargs,
    ) -> Any:
        """Async version of the request method."""
        if raise_for_status is None:
            raise_for_status = method_config.raise_for_status

        logger.debug(
            'Request url asynchronously',
            extra={
                'method_config': method_config.model_dump(exclude_none=True),
                'raise_for_status': raise_for_status,
                'kwargs': kwargs,
            },
        )

        custom_params = {key: val for key, val in kwargs.pop('params', {}).items() if val is not None}
        custom_url, custom_url_params = extract_url_params(kwargs.pop('url', None))
        custom_method = kwargs.pop('method', None)

        if self.base_url:
            method_config.base_url = self.base_url

        httpx_response = await self.asynchronous.request(
            method=custom_method or method_config.request_verb.value,
            url=custom_url or method_config.url(**kwargs),
            params=custom_params | custom_url_params or None,
            **extract_method_params(httpx.request, kwargs),
        )
        response = Response(
            httpx_response=httpx_response,
            json_path=method_config.json_path,
            raise_for_status=raise_for_status,
            response_model=method_config.response_model,
        )

        if filepath and method_config.save_method:
            await response.async_save_data(filepath)
            return None
        return response.data
